@@ -43,9 +43,17 @@ from .utils import validate
 from .utils import LazyString
 from . import __version__
 
-NO_SANITIZER = lambda text: text  # noqa
-BR_SANITIZER = lambda text: text.replace('\n', '<br/>') if text else text  # noqa
-MK_SANITIZER = lambda text: Markup(markdown(text)) if text else text  # noqa
+
+def NO_SANITIZER(text):
+    return text
+
+
+def BR_SANITIZER(text):
+    return text.replace('\n', '<br/>') if text else text
+
+
+def MK_SANITIZER(text):
+    return Markup(markdown(text)) if text else text
 
 
 class APIDocsView(MethodView):
@@ -87,6 +95,23 @@ class APIDocsView(MethodView):
                 'favicon',
                 url_for('flasgger.static', filename='favicon-32x32.png')
             )
+            data['swagger_ui_bundle_js'] = self.config.get(
+                'swagger_ui_bundle_js',
+                url_for('flasgger.static', filename='swagger-ui-bundle.js')
+            )
+            data['swagger_ui_standalone_preset_js'] = self.config.get(
+                'swagger_ui_standalone_preset_js',
+                url_for('flasgger.static',
+                        filename='swagger-ui-standalone-preset.js')
+            )
+            data['jquery_js'] = self.config.get(
+                'jquery_js',
+                url_for('flasgger.static', filename='lib/jquery.min.js')
+            )
+            data['swagger_ui_css'] = self.config.get(
+                'swagger_ui_css',
+                url_for('flasgger.static', filename='swagger-ui.css')
+            )
             return render_template(
                 'flasgger/index.html',
                 **data
@@ -137,8 +162,8 @@ class Swagger(object):
         "specs_route": "/apidocs/"
     }
 
-    SCHEMA_TYPES = {'string': str, 'integer': int, 'number': int,
-                    'boolean': bool, 'object': dict}
+    SCHEMA_TYPES = {'string': str, 'integer': int, 'number': float,
+                    'boolean': bool}
     SCHEMA_LOCATIONS = {'query': 'args', 'header': 'headers',
                         'formData': 'form', 'body': 'json', 'path': 'path'}
 
@@ -196,7 +221,7 @@ class Swagger(object):
         if filename.endswith('.json'):
             loader = json.load
         elif filename.endswith('.yml') or filename.endswith('.yaml'):
-            loader = yaml.load
+            loader = yaml.safe_load
         else:
             with codecs.open(filename, 'r', 'utf-8') as f:
                 contents = f.read()
@@ -204,7 +229,7 @@ class Swagger(object):
                 if contents[0] in ['{', '[']:
                     loader = json.load
                 else:
-                    loader = yaml.load
+                    loader = yaml.safe_load
         with codecs.open(filename, 'r', 'utf-8') as f:
             return loader(f)
 
@@ -238,7 +263,7 @@ class Swagger(object):
         }
 
     def get_apispecs(self, endpoint='apispec_1'):
-        if endpoint in self.apispecs:
+        if not self.app.debug and endpoint in self.apispecs:
             return self.apispecs[endpoint]
 
         spec = None
@@ -334,7 +359,11 @@ class Swagger(object):
         for rule, verbs in specs:
             operations = dict()
             for verb, swag in verbs:
-                definitions.update(swag.get('definitions', {}))
+                update_dict = swag.get('definitions', {})
+                if type(update_dict) == list and type(update_dict[0]) == dict:
+                    # pop, assert single element
+                    update_dict, = update_dict
+                definitions.update(update_dict)
                 defs = []  # swag.get('definitions', [])
                 defs += extract_definitions(
                     defs, endpoint=rule.endpoint, verb=verb,
@@ -362,6 +391,19 @@ class Swagger(object):
                         prefix_ids=prefix_ids
                     )
 
+                callbacks = swag.get("callbacks", {})
+                if callbacks:
+                    callbacks = {
+                        str(key): value
+                        for key, value in callbacks.items()
+                    }
+                    extract_definitions(
+                        list(callbacks.values()),
+                        endpoint=rule.endpoint,
+                        verb=verb,
+                        prefix_ids=prefix_ids
+                    )
+
                 responses = None
                 if 'responses' in swag:
                     responses = swag.get('responses', {})
@@ -391,6 +433,8 @@ class Swagger(object):
                     operation['description'] = swag.get('description')
                 if request_body:
                     operation['requestBody'] = request_body
+                if callbacks:
+                    operation['callbacks'] = callbacks
                 if responses:
                     operation['responses'] = responses
                 # parameters - swagger ui dislikes empty parameter lists
@@ -519,6 +563,10 @@ class Swagger(object):
     def parse_request(self, app):
         @app.before_request
         def before_request():  # noqa
+            """
+            Parse and validate request data(query, form, header and body),
+            set data to `request.parsed_data`
+            """
             # convert "/api/items/<int:id>/" to "/api/items/{id}/"
             subs = []
             for sub in str(request.url_rule).split('/'):
@@ -533,7 +581,7 @@ class Swagger(object):
             path = '/'.join(subs)
             path_key = path + request.method.lower()
 
-            if path_key in self.parsers:
+            if not self.app.debug and path_key in self.parsers:
                 parsers = self.parsers[path_key]
                 schemas = self.schemas[path_key]
             else:
@@ -552,21 +600,10 @@ class Swagger(object):
                 schemas = defaultdict(
                     lambda: {'type': 'object', 'properties': defaultdict(dict)}
                 )
-                for param in doc['parameters']:
+                for param in doc.get('parameters', []):
                     location = self.SCHEMA_LOCATIONS[param['in']]
-                    if location == 'json':
-                        schemas[location]['properties'].update(
-                            param['schema']['properties'])
-
-                        required_keys = param['schema'].get('required', [])
-                        keys = param['schema']['properties']
-                        for key in keys:
-                            parsers[location].add_argument(
-                                key,
-                                type=self.SCHEMA_TYPES[
-                                    keys[key]['type']],
-                                required=key in required_keys, location='json',
-                                store_missing=False)
+                    if location == 'json':  # load data from 'request.json'
+                        schemas[location] = param['schema']
                     else:
                         name = param['name']
                         if location != 'path':
@@ -590,6 +627,8 @@ class Swagger(object):
             parsed_data = {'path': request.view_args}
             for location in parsers.keys():
                 parsed_data[location] = parsers[location].parse_args()
+            if 'json' in schemas:
+                parsed_data['json'] = request.json or {}
             for location, data in parsed_data.items():
                 try:
                     jsonschema.validate(
